from typing import override

from ai.backend.test.contexts.context import BaseTestContext, ContextName
from ai.backend.test.data.session import CreatedSessionMeta, SessionDependency
from ai.backend.test.tester.dependency import (
    BatchSessionDep,
    BootstrapScriptDep,
    ClusterDep,
    CodeExecutionDep,
    SessionDep,
    SessionImagifyDep,
)


class SessionContext(BaseTestContext[SessionDep]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.SESSION


class BatchSessionContext(BaseTestContext[BatchSessionDep]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.BATCH_SESSION


class BootstrapScriptContext(BaseTestContext[BootstrapScriptDep]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.BOOTSTRAP_SCRIPT


class ClusterContext(BaseTestContext[ClusterDep]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CLUSTER_CONFIG


class CodeExecutionContext(BaseTestContext[CodeExecutionDep]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CODE_EXECUTION


class CreatedSessionTemplateIDContext(BaseTestContext[str]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CREATED_SESSION_TEMPLATE_ID


class CreatedSessionMetaContext(BaseTestContext[CreatedSessionMeta]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CREATED_SESSION_META


class SessionDependencyContext(BaseTestContext[SessionDependency]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.SESSION_DEPENDENCY


class SessionImagifyContext(BaseTestContext[SessionImagifyDep]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.SESSION_IMAGIFY



import asyncio
from typing import override

import aiohttp

from ai.backend.test.contexts.model_service import (
    CreatedModelServiceEndpointMetaContext,
    CreatedModelServiceTokenContext,
)
from ai.backend.test.templates.template import TestCode

_ENDPOINT_HEALTH_CHECK_TIMEOUT = 20


# TODO: Consider health check points for each runtime variant (See MODEL_SERVICE_RUNTIME_PROFILES)
def _make_endpoint_health_check_url(endpoint_url: str) -> str:
    """
    Constructs the health check URL for the given endpoint URL.
    """
    return f"{endpoint_url}/health"


class EndpointHealthCheck(TestCode):
    _expected_status_codes: set[int]

    def __init__(self, expected_status_codes: set[int]) -> None:
        super().__init__()
        self._expected_status_codes = expected_status_codes

    @override
    async def test(self) -> None:
        service_meta = CreatedModelServiceEndpointMetaContext.current()
        health_check_url = _make_endpoint_health_check_url(service_meta.endpoint_url)

        async with aiohttp.ClientSession() as http_sess:
            resp = await asyncio.wait_for(
                http_sess.get(health_check_url), timeout=_ENDPOINT_HEALTH_CHECK_TIMEOUT
            )

            assert resp.status in self._expected_status_codes, (
                f"Service endpoint health check failed with status: {resp.status}, expected {self._expected_status_codes}"
            )


class EndpointHealthCheckWithToken(TestCode):
    _expected_status_codes: set[int]

    def __init__(self, expected_status_codes: set[int]) -> None:
        super().__init__()
        self._expected_status_codes = expected_status_codes

    @override
    async def test(self) -> None:
        service_meta = CreatedModelServiceEndpointMetaContext.current()
        health_check_url = _make_endpoint_health_check_url(service_meta.endpoint_url)

        token = CreatedModelServiceTokenContext.current()
        headers = {"Authorization": f"Bearer {token}"}

        async with aiohttp.ClientSession() as http_sess:
            resp = await asyncio.wait_for(
                http_sess.get(health_check_url, headers=headers),
                timeout=_ENDPOINT_HEALTH_CHECK_TIMEOUT,
            )

            assert resp.status in self._expected_status_codes, (
                f"Service endpoint health check failed with status: {resp.status}, expected {self._expected_status_codes}"
            )



"""
An integration test for the Backend.AI admin APIs.

It runs and checks the result of a series of CRUD commands for various entities including domain, group,
user, scaling group, resource policy, resource preset, etc.

The location of the client executable and the credential scripts are configured as environment variables
read by the pytest fixtures in the ai.backend.test.cli_integration.conftest module.
"""



from typing import Any, Optional

from pydantic import BaseModel, ConfigDict, Field, model_validator

from ai.backend.common.types import ClusterMode, RuntimeVariant


class BaseDependencyModel(BaseModel):
    @staticmethod
    def snake_to_kebab_case(string: str) -> str:
        return string.replace("_", "-")

    model_config = ConfigDict(
        populate_by_name=True,
        from_attributes=True,
        use_enum_values=True,
        extra="allow",
        alias_generator=snake_to_kebab_case,
    )


class KeyPairDep(BaseDependencyModel):
    access_key: str = Field(description="The access key for the API.", examples=["some-access-key"])
    secret_key: str = Field(description="The secret key for the API.", examples=["some-secret-key"])


class UserResourcePolicyDep(BaseDependencyModel):
    name: str = Field(
        description="The name of the user resource policy.",
        examples=["default"],
    )


class EndpointDep(BaseDependencyModel):
    api_endpoint: Optional[str] = Field(
        default=None,
        description="The API endpoint configuration for the test context.",
    )
    login_endpoint: Optional[str] = Field(
        default=None,
        description="The login endpoint configuration for the test context.",
    )


class LoginCredentialDep(BaseDependencyModel):
    user_id: str = Field(description="The user ID for login.", examples=["user@example.com"])
    password: str = Field(description="The password for login.", examples=["password123"])
    otp: Optional[str] = Field(
        default=None,
        description="The one-time password for login, if required.",
        examples=["123456"],
    )


class DomainDep(BaseDependencyModel):
    name: str = Field(
        description="The domain name for the test context.",
        examples=["default"],
    )


class GroupDep(BaseDependencyModel):
    name: str = Field(
        description="The group name for the test context.",
        examples=["default"],
    )


class ScalingGroupDep(BaseDependencyModel):
    name: str = Field(
        description="The scaling group for the test context.",
        examples=["default"],
    )


class ImageDep(BaseDependencyModel):
    name: str = Field(
        description="The Docker image to use for the test context.",
        examples=["cr.backend.ai/multiarch/python:3.13-ubuntu24.04"],
    )
    architecture: str = Field(
        description="The architecture of the session.",
        examples=["x86_64"],
    )


class BatchSessionDep(BaseDependencyModel):
    startup_command: str = Field(
        description="The startup command to run in the batch session.",
        examples=["ls -la"],
    )
    batch_timeout: Optional[float] = Field(
        default=None,
        description="The timeout for the batch session in seconds.",
        examples=[10.0],
    )


class SSEDep(BaseDependencyModel):
    timeout: float = Field(
        default=60.0,
        description="The timeout for the session creation in seconds.",
        examples=[60.0],
    )


class ClusterDep(BaseDependencyModel):
    # By default, testing is conducted for both single-node and multi-node setups through parametrization,
    # But we'd like to have left room for manually injecting values.
    cluster_mode: ClusterMode = Field(
        description="The cluster mode for the session.",
        examples=["single-node", "multi-node"],
    )
    cluster_size: int = Field(
        description="The size of the cluster for the session.",
        examples=[1, 2, 3],
    )


class BootstrapScriptDep(BaseDependencyModel):
    bootstrap_script: str = Field(
        description="The bootstrap script to run before the session starts. Used as an argument when creating a compute session.",
        examples=["echo 'Bootstrapping...'"],
    )


class SessionDep(BaseDependencyModel):
    resources: Optional[dict[str, Any]] = Field(
        default=None,
        description="The resources to allocate for the session.",
        examples=[{"cpu": 2, "mem": "4gb"}],
    )


class SessionImagifyDep(BaseDependencyModel):
    new_image_name: str = Field(
        description="The name of the new image to create from the session.",
        examples=["my-custom-image"],
    )


class ModelServiceDep(BaseDependencyModel):
    model_vfolder_name: str = Field(
        description="The model VFolder name to use for the model service.",
        examples=["vfolder-name"],
    )
    replicas: int = Field(
        description="The number of replicas for the model service.",
        examples=[1, 2, 3],
    )
    runtime_variant: RuntimeVariant = Field(
        default=RuntimeVariant.CUSTOM,
        description="The runtime variant for the model service.",
        examples=[v.name for v in RuntimeVariant],
    )
    # Separate group is required for the model service, so we placed this independently from the group context.
    group_name: str = Field(
        description="The group name for the model service.",
        examples=["model-store"],
    )
    model_mount_destination: str = Field(
        default="models",
        description="The destination path for the model mount in the model service.",
        examples=["models"],
    )
    model_definition_path: str = Field(
        default="./model-definition.yaml",
        description="The path to the model definition file in the model service.",
        examples=["./model-definition.yaml"],
    )


class VFolderDep(BaseDependencyModel):
    group: Optional[str] = Field(
        default=None,
        description="The group name for the vfolder.",
    )
    unmanaged_path: Optional[str] = Field(
        default=None, description="The unmanaged path for the vfolder."
    )
    permission: str = Field(
        description="The permission for the vfolder.",
    )
    cloneable: bool = Field(
        description="Whether the vfolder is cloneable.",
    )
    share_permission: str = Field(
        default="ro",
        description="The share permission for the vfolder.",
        examples=["ro", "rw"],
    )

    @model_validator(mode="after")
    def validate(self):
        if self.permission == self.share_permission:
            raise ValueError("permission and share_permission must be different")
        return self


class UploadFileDep(BaseDependencyModel):
    path: str = Field(
        description="The name of the file to upload.",
        examples=["test_file.txt", "nested/inner_file.txt"],
    )
    content: str = Field(
        description="The content of the file to upload.",
        examples=["This is a test file content."],
    )


class TestContextInjectionModel(BaseDependencyModel):
    endpoint: Optional[EndpointDep] = Field(
        default=None,
        description="The endpoint configurations for the test context.",
    )
    keypair: Optional[KeyPairDep] = Field(
        default=None,
        description="The key pair for the test context.",
    )
    login_credential: Optional[LoginCredentialDep] = Field(
        default=None,
        description="The login credentials for the test context.",
    )
    domain: Optional[DomainDep] = Field(
        default=None,
        description="The domain configuration for the test context.",
    )
    group: Optional[GroupDep] = Field(
        default=None,
        description="The group configuration for the test context.",
    )
    scaling_group: Optional[ScalingGroupDep] = Field(
        default=None,
        description="The scaling group configuration for the test context.",
    )
    image: Optional[ImageDep] = Field(
        default=None,
        description="The Docker image context for the test.",
    )
    sse: Optional[SSEDep] = Field(
        default=None,
        description="The Server-Sent Events configuration for the test context.",
    )
    cluster_config: Optional[ClusterDep] = Field(
        default=None,
        description="The cluster configuration for the test context.",
    )
    batch_session: Optional[BatchSessionDep] = Field(
        default=None,
        description="The batch session configuration for the test context.",
    )
    session: Optional[SessionDep] = Field(
        default=None,
        description="The session configuration for the test context.",
    )
    session_imagify: Optional[SessionImagifyDep] = Field(
        default=None,
        description="The session imagify configuration for the test context.",
    )
    model_service: Optional[ModelServiceDep] = Field(
        default=None,
        description="The model service configuration for the test context.",
    )
    vfolder: Optional[VFolderDep] = Field(
        default=None,
        description="The vfolder configuration for the test context.",
    )
    user_resource_policy: Optional[UserResourcePolicyDep] = Field(
        default=None,
        description="The user resource policy configuration for the test context.",
    )


class CodeExecutionDep(BaseDependencyModel):
    code: str = Field(
        description="The code to execute in the test.",
        examples=["print('Hello, World!')"],
    )
    expected_result: str = Field(
        description="The expected result of the code execution.",
        examples=["Hello, World!"],
    )



import textwrap

from ai.backend.test.contexts.context import ContextName
from ai.backend.test.templates.auth.keypair import (
    KeypairAuthAsCreatedUserTemplate,
    KeypairAuthTemplate,
)
from ai.backend.test.templates.template import BasicTestTemplate
from ai.backend.test.templates.user.user import UserTemplate
from ai.backend.test.templates.vfolder.file_uploader import PlainTextFilesUploader
from ai.backend.test.templates.vfolder.general_vfolder import (
    ProjectVFolderTemplate,
    UserVFolderTemplate,
)
from ai.backend.test.templates.vfolder.invite import (
    AcceptInvitationTemplate,
    RejectInvitationTemplate,
    VFolderInviteTemplate,
)
from ai.backend.test.templates.vfolder.share import ShareVFolderTemplate
from ai.backend.test.testcases.spec_manager import TestSpec, TestTag
from ai.backend.test.testcases.vfolder.access import VFolderAccessFailure, VFolderAccessSuccess
from ai.backend.test.testcases.vfolder.clone import VFolderCloneSuccess
from ai.backend.test.testcases.vfolder.delete_files import (
    VFolderFilesDeletionSuccess,
    VFolderFilesRecursiveDeletionSuccess,
)
from ai.backend.test.testcases.vfolder.download import VFolderDownloadSuccess
from ai.backend.test.testcases.vfolder.invite_failures import (
    VFolderAcceptDuplicatedInvitation,
    VFolderInviteFailure,
)
from ai.backend.test.testcases.vfolder.list_files import VFolderListFilesSuccess
from ai.backend.test.testcases.vfolder.move import VFolderFileMoveSuccess
from ai.backend.test.testcases.vfolder.rename_file import VFolderFileRenameSuccess
from ai.backend.test.testcases.vfolder.share import VFolderSharePermissionOverrideSuccess
from ai.backend.test.tester.dependency import UploadFileDep

_TEST_FILE_CONTENT = "This is a test file for VFolder download."

VFOLDER_INVITATION_TEST_SPECS = {
    "invite_vfolder_success": TestSpec(
        name="invite_vfolder_success",
        description=textwrap.dedent("""\
            Test for VFolder invitation functionality.
            The test will:
            1. Authenticate as admin and create a VFolder.
            2. Create a test user.
            3. Invite the test user to the VFolder.
            4. Login as the test user.
            5. Accept the VFolder invitation.
            6. Verify that the test user can access the VFolder.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderAccessSuccess()).with_wrappers(
            # Run as admin
            KeypairAuthTemplate,
            UserVFolderTemplate,
            UserTemplate,
            VFolderInviteTemplate,
            # Run as invitee user
            KeypairAuthAsCreatedUserTemplate,
            AcceptInvitationTemplate,
        ),
        parametrizes={
            ContextName.VFOLDER_INVITATION_PERMISSION: ["rw", "ro"],
        },
    ),
    "invite_vfolder_failure_reject": TestSpec(
        name="invite_vfolder_failure_reject",
        description=textwrap.dedent("""\
            Test for VFolder invitation functionality.
            The test will:
            1. Authenticate as admin and create a VFolder.
            2. Create a test user.
            3. Invite the test user to the VFolder.
            4. Login as the test user.
            5. Reject the VFolder invitation.
            6. Verify that the test user cannot access the VFolder.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderAccessFailure()).with_wrappers(
            # Run as admin
            KeypairAuthTemplate,
            UserVFolderTemplate,
            UserTemplate,
            VFolderInviteTemplate,
            # Run as invitee user
            KeypairAuthAsCreatedUserTemplate,
            RejectInvitationTemplate,
        ),
        parametrizes={
            ContextName.VFOLDER_INVITATION_PERMISSION: ["rw", "ro"],
        },
    ),
    "invite_vfolder_failure_duplicated_accept": TestSpec(
        name="invite_vfolder_failure_duplicated_accept",
        description=textwrap.dedent("""\
            Test for VFolder invitation functionality.
            The test will:
            1. Authenticate as admin and create a VFolder.
            2. Create a test user.
            3. Invite the test user to the VFolder.
            4. Login as the test user.
            5. Accept the VFolder invitation.
            6. Try to accept the same invitation again.
            7. Verify that the second acceptance fails with an error.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderAcceptDuplicatedInvitation()).with_wrappers(
            # Run as admin
            KeypairAuthTemplate,
            UserVFolderTemplate,
            UserTemplate,
            VFolderInviteTemplate,
            # Run as invitee user
            KeypairAuthAsCreatedUserTemplate,
        ),
        parametrizes={
            ContextName.VFOLDER_INVITATION_PERMISSION: ["rw"],
        },
    ),
    "invite_vfolder_failure_duplicated_invitation": TestSpec(
        name="invite_vfolder_failure_duplicated_invitation",
        description=textwrap.dedent("""\
            Test for VFolder invitation functionality.
            The test will:
            1. Authenticate as admin and create a VFolder.
            2. Create a test user.
            3. Invite the test user to the VFolder.
            4. Login as the test user.
            5. Accept the VFolder invitation.
            6. Try to invite the same user again to the same VFolder.
            7. Verify that the second invitation fails with an error.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderInviteFailure()).with_wrappers(
            # Run as admin
            KeypairAuthTemplate,
            UserVFolderTemplate,
            UserTemplate,
            VFolderInviteTemplate,
            # Run as invitee user
            KeypairAuthAsCreatedUserTemplate,
            AcceptInvitationTemplate,
        ),
        parametrizes={
            ContextName.VFOLDER_INVITATION_PERMISSION: ["rw"],
        },
    ),
}

VFOLDER_TEST_SPECS = {
    "list_files": TestSpec(
        name="list_files",
        description=textwrap.dedent("""\
            Test for renaming a file in a VFolder.
            The test will:
            1. Create a VFolder.
            2. Upload files to the VFolder.
            3. List files in the VFolder.
            4. Verify that the files are listed correctly.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderListFilesSuccess()).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate, PlainTextFilesUploader
        ),
        parametrizes={
            ContextName.VFOLDER_UPLOAD_FILES: [
                [
                    UploadFileDep(
                        path="test_1.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="test_2.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="test_3.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                ]
            ]
        },
    ),
    "rename_file": TestSpec(
        name="rename_file",
        description=textwrap.dedent("""\
            Test for renaming a file in a VFolder.
            The test will:
            1. Create a VFolder.
            2. Upload files to the VFolder.
            3. Rename a file in the VFolder.
            4. Verify that the file has been renamed successfully.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderFileRenameSuccess()).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate, PlainTextFilesUploader
        ),
        parametrizes={
            ContextName.VFOLDER_UPLOAD_FILES: [
                [
                    UploadFileDep(
                        path="file.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                ]
            ]
        },
    ),
    "move_file": TestSpec(
        name="move_file",
        description=textwrap.dedent("""\
            Test for renaming a file in a VFolder.
            The test will:
            1. Create a VFolder.
            2. Upload files to the VFolder.
            3. Move a file in the VFolder.
            4. Verify that the file has been moved successfully.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderFileMoveSuccess()).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate, PlainTextFilesUploader
        ),
        parametrizes={
            ContextName.VFOLDER_UPLOAD_FILES: [
                [
                    UploadFileDep(
                        path="file.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                ]
            ]
        },
    ),
    "delete_files": TestSpec(
        name="delete_files",
        description=textwrap.dedent("""\
            Test for renaming a file in a VFolder.
            The test will:
            1. Create a VFolder.
            2. Upload files to the VFolder.
            3. Delete some files in the VFolder.
            4. Verify that the files have been deleted successfully.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderFilesDeletionSuccess()).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate, PlainTextFilesUploader
        ),
        parametrizes={
            ContextName.VFOLDER_UPLOAD_FILES: [
                [
                    UploadFileDep(
                        path="test_1.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="test_2.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="test_3.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                ]
            ]
        },
    ),
    "delete_files_recursively": TestSpec(
        name="delete_files_recursively",
        description=textwrap.dedent("""\
            Test for renaming a file in a VFolder.
            The test will:
            1. Create a VFolder.
            2. Upload files to the VFolder.
            3. Delete some files in the VFolder recursively.
            4. Verify that the files have been deleted successfully.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderFilesRecursiveDeletionSuccess("nested")).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate, PlainTextFilesUploader
        ),
        parametrizes={
            ContextName.VFOLDER_UPLOAD_FILES: [
                [
                    UploadFileDep(
                        path="nested/test_1.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="nested/test_2.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="nested/inner_nested/test_3.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                ]
            ]
        },
    ),
    "download_files": TestSpec(
        name="download_files",
        description=textwrap.dedent("""\
            Test for downloading files in a VFolder.
            The test will:
            1. Create a VFolder.
            2. Upload files to the VFolder.
            3. Download the files from the VFolder.
            4. Verify that the downloaded files match the uploaded files.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderDownloadSuccess()).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate, PlainTextFilesUploader
        ),
        parametrizes={
            ContextName.VFOLDER_UPLOAD_FILES: [
                [
                    UploadFileDep(
                        path="test_1.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="test_2.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="test_3.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="nested/inner.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                ]
            ]
        },
    ),
    "clone_vfolder": TestSpec(
        name="clone_vfolder",
        description=textwrap.dedent("""\
            Test for renaming a file in a VFolder.
            The test will:
            1. Create a VFolder.
            2. Upload files to the VFolder.
            3. Clone the VFolder.
            4. Verify that the files in the cloned VFolder match the original VFolder.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderCloneSuccess()).with_wrappers(
            KeypairAuthTemplate, ProjectVFolderTemplate, PlainTextFilesUploader
        ),
        parametrizes={
            ContextName.VFOLDER_UPLOAD_FILES: [
                [
                    UploadFileDep(
                        path="test_1.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="test_2.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                    UploadFileDep(
                        path="nested/test_3.txt",
                        content=_TEST_FILE_CONTENT,
                    ),
                ]
            ]
        },
    ),
    "shared_vfolder_permission_override": TestSpec(
        name="shared_vfolder_permission_override",
        description=textwrap.dedent("""\
            Test for VFolder sharing functionality.
            The test will:
            1. Authenticate as admin and create a VFolder.
            2. Create a test user.
            3. Share the VFolder with the test user.
            4. Login as the test user.
            5. Verify that the test user can access the VFolder with the shared permission.
            6. Unshare the VFolder from the test user.
        """),
        tags={TestTag.MANAGER, TestTag.VFOLDER},
        template=BasicTestTemplate(VFolderSharePermissionOverrideSuccess()).with_wrappers(
            # Run as admin(sharing user)
            KeypairAuthTemplate,
            ProjectVFolderTemplate,
            UserTemplate,
            ShareVFolderTemplate,
            # Run as shared user
            KeypairAuthAsCreatedUserTemplate,
        ),
    ),
    **VFOLDER_INVITATION_TEST_SPECS,
}



from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext
from ai.backend.test.templates.template import TestCode


class VFolderDeleteAndRestoreSuccess(TestCode):
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()

        try:
            await client_session.VFolder(str(vfolder_meta.name)).delete()
        except Exception as e:
            raise AssertionError(f"Failed to delete vfolder: {e}")

        vfolder_info = await client_session.VFolder(str(vfolder_meta.name)).info()
        assert vfolder_info["status"] == "delete-pending", (
            "VFolder is not soft-deleted successfully"
        )

        try:
            await client_session.VFolder(str(vfolder_meta.name)).restore()
        except Exception as e:
            raise AssertionError(f"Failed to restore vfolder: {e}")

        vfolder_info = await client_session.VFolder(str(vfolder_meta.name)).info()
        assert vfolder_info["status"] == "ready", "VFolder is not restored successfully"



import os
import re
import shutil
from dataclasses import dataclass
from typing import override

from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.session import CreatedSessionMetaContext
from ai.backend.test.contexts.tester import TestSpecMetaContext
from ai.backend.test.contexts.vfolder import CreatedVFolderMetaContext
from ai.backend.test.templates.template import TestCode


@dataclass
class _UploadFileMeta:
    file_name: str
    full_path: str
    base_dir: str
    content: str


class FileHandlingInMountedVFolderSuccess(TestCode):
    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        vfolder_meta = CreatedVFolderMetaContext.current()
        session_meta = CreatedSessionMetaContext.current()
        session_name = session_meta.name

        try:
            file_upload_meta = await self._create_dummy_file()

            await client_session.VFolder(name=vfolder_meta.name).upload(
                sources=[file_upload_meta.file_name],
                basedir=file_upload_meta.base_dir,
            )
            exec_result = await client_session.ComputeSession(name=session_name).list_files(
                path=vfolder_meta.name
            )
            self._verify_file_exists(exec_result, file_upload_meta.file_name)
            await client_session.ComputeSession(name=session_name).download(
                files=[f"./{vfolder_meta.name}/{file_upload_meta.file_name}"],
                dest=f"{file_upload_meta.base_dir}/downloaded",
            )
            self._verify_downloaded_file_identical(
                original_file_path=file_upload_meta.full_path,
                downloaded_file_path=f"{file_upload_meta.base_dir}/downloaded/{file_upload_meta.file_name}",
            )
        finally:
            if file_upload_meta:
                await self._cleanup_dummy_file(file_upload_meta)

    def _verify_downloaded_file_identical(
        self, original_file_path: str, downloaded_file_path: str
    ) -> None:
        with open(original_file_path, "r", encoding="utf-8") as original_file:
            original_content = original_file.read()

        with open(downloaded_file_path, "r", encoding="utf-8") as downloaded_file:
            downloaded_content = downloaded_file.read()

        assert original_content == downloaded_content, (
            f"Downloaded file content does not match original. "
            f"Original: {original_content}, Downloaded: {downloaded_content}"
        )

    def _verify_file_exists(self, files_result: dict, original_filename: str) -> None:
        files_str = files_result.get("files", "")

        filename_pattern = r'"filename":\s*"([^"]+)"'
        matches = re.findall(filename_pattern, files_str)

        for filename in matches:
            if filename == original_filename:
                return

        raise AssertionError(f"File {original_filename} not found in the files string: {files_str}")

    async def _create_dummy_file(self) -> _UploadFileMeta:
        test_spec = TestSpecMetaContext.current()
        current_dir = os.getcwd()
        base_dir = os.path.join(current_dir, f"test_{test_spec.test_id}")

        os.makedirs(base_dir, exist_ok=True)

        file_name = f"{test_spec.test_id}.txt"
        full_path = os.path.join(base_dir, file_name)
        file_content = f"This is a dummy file for test {test_spec.test_id}."

        with open(full_path, "w", encoding="utf-8") as f:
            f.write(file_content)

        return _UploadFileMeta(
            file_name=file_name, full_path=full_path, base_dir=base_dir, content=file_content
        )

    async def _cleanup_dummy_file(self, file_meta: _UploadFileMeta) -> None:
        if os.path.exists(file_meta.full_path):
            os.remove(file_meta.full_path)
        if os.path.exists(file_meta.base_dir):
            shutil.rmtree(file_meta.base_dir)



from typing import override

from ai.backend.client.exceptions import BackendAPIError
from ai.backend.test.contexts.client_session import ClientSessionContext
from ai.backend.test.contexts.model_service import (
    CreatedModelServiceEndpointMetaContext,
)
from ai.backend.test.contexts.resource_policy import UserResourcePolicyContext
from ai.backend.test.templates.template import TestCode
from ai.backend.test.utils.exceptions import UnexpectedSuccess


class ReplicasScaleFailureTooManySessions(TestCode):
    """
    Test case to verify that scaling replicas beyond the maximum allowed sessions fails
    """

    @override
    async def test(self) -> None:
        client_session = ClientSessionContext.current()
        model_service_endpoint_meta = CreatedModelServiceEndpointMetaContext.current()
        service_id = model_service_endpoint_meta.service_id

        user_resource_policy_dep = UserResourcePolicyContext.current()
        resource_policy_info = await client_session.UserResourcePolicy(
            name=user_resource_policy_dep.name
        ).get_info()
        max_session_count_per_model_service = resource_policy_info[
            "max_session_count_per_model_session"
        ]

        try:
            await client_session.Service(id=service_id).scale(
                max_session_count_per_model_service + 1
            )
            raise UnexpectedSuccess(
                "Expected BackendAPIError for overscaling replicas was not raised"
            )
        except BackendAPIError as e:
            assert e.status == 400
            assert e.data["error_code"] == "api_parsing_invalid-parameters"



from typing import override

from ai.backend.test.contexts.context import BaseTestContext, ContextName
from ai.backend.test.data.user import CreatedUserMeta


class CreatedUserContext(BaseTestContext[CreatedUserMeta]):
    @override
    @classmethod
    def name(cls) -> ContextName:
        return ContextName.CREATED_USER_CONTEXT
